var gPromptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                               .getService(Components.interfaces.nsIPromptService);

var gAtomService = Components.classes["@mozilla.org/atom-service;1"].getService(Components.interfaces.nsIAtomService);
var gAtoms = {
  separator: gAtomService.getAtom("separator")
}

var gDocument, gLocation, gPref, gData, gItem, gApply, gIsClosing;
var gFallbacks = 1;
var devBox;

var resettedItems;

var menuTree, attrTree;

function init(target) {
  if(!target) return;

  var containers, l, i;

  gData = []; resettedItems = []; gApply = false;
  gDocument = target.document;
  gLocation = gDocument.location.href;
  gPref = target.menumanipulator.ps;

  document.title = gStrings.title + " - " + gLocation;

  menuTree = document.getElementById("menu-tree");
  attrTree = document.getElementById("attr-tree");
  devBox = document.getElementById("devbox");

  containers = gDocument.getElementsByTagName("menubar");
  for(i = 0, l = containers.length; i < l; i++) {
    gData.push({label:"menubar",node:containers[i],target:containers[i],level:0,container:true,open:true,changes:{set:[]}});
    gData = gData.concat(addData(containers[i].childNodes,1));
  }

  containers = gDocument.getElementsByTagName("popup");
  for(i = 0, l = containers.length; i < l; i++) gData = gData.concat(addData([containers[i]],0));

  var tabContext = gDocument.getElementById("menumanipulator-tabcontext");
  if(tabContext) gData = gData.concat(addData([tabContext],0));

  menuTree.view.selection.select(-1);
  menuTree.view = menuView;

  attrView.data = { length:0 };
  attrTree.view = attrView;

  if(gPref.getPrefType(gLocation))
    devBox.value = gPref.getCharPref(gLocation).replace(/;   /g, ";\n\n").replace(/; /g, ";\n");
  else
    devBox.value = "";

  gFallbacks = gPref.getIntPref("Move.fallbacks");
}

function addData(nodes,level) {
  var retval = [];
  for(var i = 0, l = nodes.length; i < l; i++) {
    var node = nodes[i];

    if(node.nodeType != 1) continue;

    if(node.id.indexOf("rdf:") > -1
    || node.id.indexOf("NC:") > -1
    || node.id.indexOf("://") > -1
    || node.hasAttribute("toolbarindex")
    || node.hasAttribute("data")
    || node.hasAttribute("index")
    ||(node.hasAttribute("image") && node.getAttribute("image").indexOf("moz-icon:") > -1) //bug 108480
    || node.className == "tooltip"
    || node.hasAttribute("userdefined")
    || node.hasAttribute("menumanipulator")
    ) continue;

    var data = { node: node, level: level, label: node.getAttribute("label"), changes: { set: [] }}
    if(node.menumanipulator) {
      data.changes.apply = true;
      data.changes.path = node.menumanipulator.path;
      data.changes.created = node.menumanipulator.created;
      data.changes.moved = node.menumanipulator.moved;
      
      var attr;
      for(attr in node.menumanipulator.set) {
        data.changes.set[attr] = node.menumanipulator.set[attr];
      }
    }

    if(!data.label && (node.hasAttribute("command") || node.hasAttribute("observes"))) {
      var command = node.getAttribute("command") || node.getAttribute("observes");
      data.label = node.ownerDocument.getElementById(command).getAttribute("label");
    }

    if(!data.label) data.label = node.id;

    if(node.hasAttribute("image")) data.image = node.getAttribute("image");
    else {
      var listStyleImage = gDocument.defaultView.getComputedStyle(node, null).listStyleImage;
      if(listStyleImage != "none" && gDocument.defaultView.getComputedStyle(node, null).MozImageRegion == "auto")
        data.image = listStyleImage.slice(4,listStyleImage.length - 1);
    }

    retval.push(data);

    switch(node.localName) {
      case "menuitem":
        if(!data.label) retval.pop();
        break;
      case "menu":
        data.container = true; data.open = true;
        if(node.firstChild && node.firstChild.localName == "menupopup") {
          data.target = node.firstChild;
          retval = retval.concat(addData(node.firstChild.childNodes,level+1));
        }
        break;
      case "menuseparator":
        data.separator = true;
        break;
      case "menupopup": case "popup":
        data.container = true; data.open = true;
        data.target = node;
        retval = retval.concat(addData(node.childNodes,level+1));
        break;
      default:
        retval.pop();
        break;
    }
  }

  return retval;
}

function createXPath(node) {
  if(node.id) return "id('"+specEscape(node.id)+"')";

  var path = "x:" + node.localName;
  if(node.hasAttribute("command")) path += "[@command='" + specEscape(node.getAttribute("command")) +"']";
  else if(node.hasAttribute("label")) path += "[@label='" + specEscape(node.getAttribute("label")) +"']";
  else if(node.hasAttribute("observes")) path += "[@observes='" + specEscape(node.getAttribute("observes")) +"']";

  return createXPath(node.parentNode) + "/" + path;
}

var menuView = {
  get rowCount() { return gData.length; },
  get dragSession() { return nsDragAndDrop.mDragService.getCurrentSession(); },

  getRowProperties: function() {}, getColumnProperties: function(){},

  isContainer: function(index) { return gData[index].container; },
  isContainerOpen: function(index) { return gData[index].open; },
  isContainerEmpty: function(index) { return false; },
  toggleOpenState: function(index) {
   var currentRow = gData[index];
   if(currentRow.open) {
    var length = 0;
    while(gData[index + ++length] &&
          gData[index].level < gData[index + length].level);
    if(length-- > 0) {
     currentRow.childs = gData.splice(index + 1,length);
     menuView.treebox.rowCountChanged(index + 1,-length);
    }
   } else {
    for(var i = 0, l = currentRow.childs.length; i < l; i++) {
     gData.splice(index + 1 + i, 0, currentRow.childs[i]);
    }
    menuView.treebox.rowCountChanged(index + 1,currentRow.childs.length);
    currentRow.childs = null;
   }

   gData[index].open = !gData[index].open;
   this.treebox.invalidateRow(index);

  },

  isSeparator: function(index) { return gData[index].separator; },
  isSorted: function() { return false; },

  canDropBeforeAfter: function(index, before) { return this.canDrop(index, before ? -1 : 1); },
  canDropOn: function(index) {return this.canDrop(index, 0);},
  canDrop: function(index, orientation) {
    if(this.dragSession.sourceNode != this.treebox.treeBody ||
       index == this.selection.currentIndex ||
       (orientation == 0 && !gData[index].target) ||
       (index == 0 && orientation == -1)
      )
    return false;

    var parentIndex = this.getParentIndex(index);
    while(parentIndex != -1) {
      if(this.selection.currentIndex == parentIndex) return false;
      parentIndex = this.getParentIndex(parentIndex);
    }

    return true;
  },
  onDragStart: function (event, transferData, action)
  {
    var item = gData[menuTree.currentIndex];
    if(item.level > 0 &&
       (item.node.localName != "menuseparator" || item.node.id) &&
       (item.changes.apply || createXPath(item.node).split("/").length < 4)
      ) {
      transferData.data = new TransferData(); transferData.data.addDataForFlavour("text/x-dummy", "");
    }
  },
  drop: function(row, orientation) {
    if(menuTree.view.inDropAfter) orientation--;

    var length = 0;
    while(gData[menuTree.currentIndex + ++length] &&
          gData[menuTree.currentIndex].level < gData[menuTree.currentIndex + length].level);

    var newIndex, levelAdjust = gData[row].level - gData[menuTree.currentIndex].level;
    if(orientation == -1) {
      newIndex = row;

      if(gData[row].level == 0) levelAdjust++;
    } else if(orientation == 0) {
      var lastChild = row;
      while(gData[++lastChild] && gData[lastChild].level > gData[row].level);
      newIndex = lastChild;

      levelAdjust++;
    } else {
      newIndex = row + 1;

      if(gData[row].target) levelAdjust++;
    }

    if(newIndex > menuTree.currentIndex) newIndex -= length;
    if(newIndex == menuTree.currentIndex && levelAdjust == 0) return;

    var currentRow = gData[menuTree.currentIndex];
    gApply = true;
    currentRow.changes.apply = true;
    currentRow.changes.move = true;

    var movedData = gData.splice(menuTree.currentIndex,length);
    var i, l;
    for(i = 0, l = movedData.length; i < l; i++) {
      gData.splice(newIndex + i, 0, movedData[i]);
      movedData[i].level += levelAdjust;
    }
 
    this.treebox.invalidate();

    if(currentRow) {
      i = -1;
      do { i++; } while(currentRow != gData[i]);
      this.selection.select(i);
      this.treebox.ensureRowIsVisible(i);
    }

  },

  getLevel: function(index) { return gData[index].level; },

  getImageSrc: function(row, col) { return gData[row].image; },

  getCellText : function(row,col){ return gData[row].label;},
  getCellProperties: function(row,col,props) {
    if(gData[row].separator) props.AppendElement(gAtoms.separator);
  },

  setTree: function(treebox) { this.treebox=treebox; },

  cycleHeader: function cycleHeader(col, elem) {  },

  hasNextSibling: function(rowIndex, afterIndex) {
    while(gData[++afterIndex] && gData[afterIndex].level >= gData[rowIndex].level)
      if(gData[afterIndex].level == gData[rowIndex].level) return afterIndex;

    return 0;
  },

  getParentIndex: function(rowIndex) { 
    var i = rowIndex;

    while(gData[--i]) if(gData[i].level < gData[rowIndex].level) return i;

    return -1;
  },

  selectionChanged: function() {
    gItem = gData[menuTree.currentIndex];

    if(menuTree.currentIndex == -1) return;

    var item = gData[menuTree.currentIndex];

    attrTree.treeBoxObject.rowCountChanged(0,-attrView.rowCount);

    var i, l;
    attrView.data = [];
    for(i = 0, l = item.node.attributes.length; i < l; i++) {
      if(item.node.attributes[i].nodeName in item.changes.set) continue;
      attrView.data.push({attrName: item.node.attributes[i].nodeName, attrValue: item.node.attributes[i].nodeValue});
    }
    var attr;
    for(attr in item.changes.set)
      attrView.data.push({attrName: attr, attrValue: item.changes.set[attr]});

    attrView.data.sort(function(a,b){ return a.attrName.localeCompare(b.attrName); });

    attrTree.treeBoxObject.rowCountChanged(0, attrView.rowCount);
  }

}

function onOK() {
  var i, l;

  i = 0;
  do {
   if(gData[i].container && !gData[i].open) menuView.toggleOpenState(i);
  } while(gData[++i]);

  var order = [], level;
  for(level = 0; level < 10; level++)
    for(i = gData.length - 1; i > -1 ; i--)
      if(gData[i].level == level && gData[i].changes.apply)
        order.push(i);

  var actions = [];
  for(i = 0, l = order.length; i < l; i++) actions.push(generateActions(gData[order[i]],order[i]));

  for(i = 0, l = resettedItems.length; i < l; i++) resettedItems[i].node.setAttribute("menumanipulator","removed");

  var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
  str.data = actions.join("  ");
  gPref.setComplexValue(gLocation, Components.interfaces.nsISupportsString, str);
}

function generateActions(item, row) {
  var retval = "";

  item.node.menumanipulator = item.changes;
  if(!item.changes.path) item.changes.path = createXPath(item.node);

  if(!item.changes.created)
    retval += 'Node("'+ item.changes.path +'"); ';

  if(item.changes.moved || item.changes.move) {
    var nextSibling = menuView.hasNextSibling(row,row);
    var relNode = nextSibling ? gData[nextSibling].node : gData[menuView.getParentIndex(row)].target;

    if(item.changes.move) {
      item.changes.moved = true;
      item.changes.move = false;
      if(nextSibling)
        relNode.parentNode.insertBefore(item.node,relNode);
      else {
        relNode.appendChild(item.node);
      }
    }

    retval += 'Move("'+ createXPath(relNode) +'", '+ (nextSibling ? -1 : -2);

    if(nextSibling) {
      var i, l, fallbacks = [];
      for(i = 0, l = gFallbacks; i < l; i++) {
        nextSibling = menuView.hasNextSibling(row,nextSibling);
        if(!nextSibling) { fallbacks.push('"' + createXPath(gData[menuView.getParentIndex(row)].node) + '",-2'); break; }
        fallbacks.push('"' + createXPath(gData[nextSibling].node) + '",-1')
      }
      if(fallbacks) retval += ",[" + fallbacks + "]";
    }

    retval += '); ';
  }
    
  var attr;

  for(attr in item.changes.set) {
    item.node.setAttribute(attr, item.changes.set[attr]);
    retval += "Set('"+specEscape(attr)+"','"+specEscape(item.changes.set[attr])+"'); ";
  }

  if(item.changes.created) {
    retval = "New('"+item.node.localName+"','"+specEscape(item.node.id)+"'); " + retval;
    if(item.node.localName == "menu")
      retval+= "New('menupopup'); Move(\"id('" + specEscape(item.node.id) + "')\",-2); ";
  }

  return retval;
}

function specEscape(val) {
 return val.replace(/\\/g,"\\\\").replace(/'/g,"\\'").replace(/\n/g,"\\n");
}

function devSave() {
  gPref.setCharPref(gLocation, devBox.value);
}

var attrView = {
  setTree: function(treebox) { this.treebox=treebox; },

  get rowCount() { return this.data.length; },

  getCellText: function(row,col){ return this.data[row][col.id || col]},

  isContainer: function() { return false; },
  isSeparator: function() { return false; },
  isSorted: function() { return false; },

  getColumnProperties: function() {},
  getCellProperties: function() {},
  getImageSrc: function() { return ""; },
  getRowProperties: function() {},

  cycleHeader: function() {},

  onEdit: function() {
    if(attrTree.currentIndex == -1) return;

    var attr = this.data[this.selection.currentIndex];
    if(!attr) return;

    var input = {value: attr.attrValue};
    if(gPromptService.prompt(window,gStrings.editTitle,gStrings.editText,input,null,{})) {
      if(attr.attrValue == input.value) return;
      attr.attrValue = input.value;
      this.treebox.invalidate();

      var item = gData[menuTree.currentIndex];
      gApply = true;
      item.changes.apply = true;
      item.changes.set[attr.attrName] = attr.attrValue;
    }
  },

  onInsert: function() {
    if(menuTree.currentIndex == -1) return;

    var attr = {}
    var input = {value: ""};

    gPromptService.prompt(window,gStrings.newTitle,gStrings.newText,input,null,{})

    attr.attrName = input.value;
    input.value = "";

    if(gPromptService.prompt(window,gStrings.newTitle,gStrings.editText,input,null,{})) {
      attr.attrValue = input.value;
      this.data.push(attr);
      this.treebox.rowCountChanged(this.data.length - 1, 1);
      var item = gData[menuTree.currentIndex];
      gApply = true;
      item.changes.apply = true;
      item.changes.set[attr.attrName] = attr.attrValue;
    }

  },

  onReset: function() {
    if(attrTree.currentIndex == -1) return;

    var attr = this.data[this.selection.currentIndex];
    var item = gData[menuTree.currentIndex];
    delete item.changes.set[attr.attrName];

    attr.attrValue = gStrings.attrOnReset;
    this.treebox.invalidate();

    gApply = true;
  }
}

function createNew(type) {
  if(gData.length == 0) return;

  var currentIndex = menuTree.currentIndex == -1 ? 0 : menuTree.currentIndex;

  var item = {
    node: gDocument.createElement(type), changes: {set: [], apply: true, move: true, created: true}
  }

  item.node.id = type+Date.now();
  item.node.menumanipulator = item.changes;

  switch(type) {
    case "menu":
      item.container = true; item.open = true;
      item.target = item.node.appendChild(document.createElement("menupopup"));
      item.changes.set["style"] = "";
    case "menuitem":
      var input = {value: ""}
      if(!gPromptService.prompt(window,gStrings.newItem.replace("$1",type),gStrings.newItemText,input,null,{})) return;

      var label = input.value;

      item.label = label;
      item.changes.set["label"] = label;
      item.node.setAttribute("label",label);
      break;
    case "menuseparator":
      item.label = item.node.id;
      item.separator = true;
      break
  }

  item.level =
    gData[currentIndex].target ? gData[currentIndex].level + 1 : gData[currentIndex].level;

  var newIndex = currentIndex + 1;
  gData.splice(newIndex,0,item);
  menuTree.treeBoxObject.rowCountChanged(newIndex,1);
  menuView.selection.select(newIndex);

  gApply = true;
}

function Reset(all) {
  var ok;
  if(all) {
    ok = confirm(gStrings.promptresetall);
    if(ok) {
      gPref.clearUserPref(gLocation);
      this.close();
    }
  } else {
    if(menuTree.currentIndex > -1) {
      gApply = true;
      var length = 0;
      while(gData[menuTree.currentIndex + ++length] &&
            gData[menuTree.currentIndex].level < gData[menuTree.currentIndex + length].level);

      resettedItems.push(gData[menuTree.currentIndex]);

      gData.splice(menuTree.currentIndex,length);
      menuView.treebox.rowCountChanged(menuTree.currentIndex,-length);
    }
  }
}

function switchWindow(event)
{
  if(gApply && !confirm(gStrings.onCancel)) return;

  var mediator = Components.classes["@mozilla.org/rdf/datasource;1?name=window-mediator"].getService();
  mediator.QueryInterface(Components.interfaces.nsIWindowDataSource);

  var target = mediator.getWindowForResource(event.target.getAttribute('id'));

  if (target) init(target);
}

function onCancel() {
  if(gApply && gIsClosing) {
    gIsClosing = false;
    return confirm(gStrings.onCancel);
  }
}